# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Enterprise Edition login dialogs."""

__all__ = ['NoticePage', 'EnterpriseRepoSetDomainPage', 'EnterpriseRepoLoginPage']

import typing
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget  # pylint: disable=no-name-in-module
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.config import AnacondaBrand
from anaconda_navigator.utils import analytics
from anaconda_navigator.widgets import ButtonPrimary
from anaconda_navigator.widgets.dialogs import StaticDialogBase
from anaconda_navigator.static.images import ANACONDA_ENTERPRISE_EDITION_LOGIN_LOGO
from . import styling
from .utils import TextContainer, USER_RE_VALIDATOR
from .base_dialogs import BaseLoginPage, BaseSettingPage


ENTERPRISE_LOGIN_TEXT_CONTAINER = TextContainer(
    info_frame_text=(
        'Login to configure Conda and Navigator to install packages from your on-premise package repository.'
    ),
    message_box_error_text='The Enterprise 4 Repository API domain is not specified! Please, set in preferences.',
    info_frame_logo_path=ANACONDA_ENTERPRISE_EDITION_LOGIN_LOGO,
)

ENTERPRISE_SET_DOMAIN_TEXT_CONTAINER = TextContainer(
    form_primary_text=(
        'Looks like this is the first time you are logging into Enterprise 4 Repository. Please set your Enterprise '
        '4 Repository API domain.'
    ),
    form_secondary_text='You only need to set this domain once. You can always change this in your Preferences.',
    form_input_label_text='Enter Enterprise 4 Repository API Domain',
    form_submit_button_text='Set Domain',
    info_frame_logo_path=ANACONDA_ENTERPRISE_EDITION_LOGIN_LOGO,
    info_frame_text=(
        'Login to configure Conda and Navigator to installpackages from your on-premise package  repository.'
    )
)


class NoticePage(StaticDialogBase):
    """
    Simple notification popup, that might be integrated into login process.

    :param title: Initial value for notification title.
    :param message: Initial value for notification message.
    :param button_text: Initial value for confirmation button text.

                        If it is not provided - login dialog will be closed ('accepted').
    """

    def __init__(
            self,
            title: str = 'You\'re almost ready!',
            message: str = 'Please verify you have the correct channels configured before proceeding',
            button_text: str = 'OK',
            parent: typing.Optional[QWidget] = None,
    ) -> None:
        """Initialize new :class:`~BaseNoticePage` instance."""
        super().__init__(parent)
        self.tracker: analytics._GATracker = analytics.GATracker()

        self._title_label = styling.LabelMainLoginTitle(title)
        self._title_label.setWordWrap(True)

        self._message_label = QLabel(message)
        self._message_label.setWordWrap(True)

        self._accept_button = ButtonPrimary(button_text)
        self._accept_button.setDefault(True)
        self._accept_button.clicked.connect(self.accept)

        self._content_layout = QVBoxLayout()
        self._content_layout.addWidget(self._message_label)
        self._content_layout.addWidget(self._accept_button, 0, Qt.AlignHCenter)

        self._content_frame = styling.WidgetNoticeFrame()
        self._content_frame.setLayout(self._content_layout)

        self._main_layout: QVBoxLayout = QVBoxLayout()
        self._main_layout.addWidget(self._title_label)
        self._main_layout.addWidget(self._content_frame)

        self.setLayout(self._main_layout)
        self.update_style_sheet()

    @property
    def title(self) -> str:  # noqa: D401
        """Title of the notification."""
        return self._title_label.text()

    @title.setter
    def title(self, value: str) -> None:
        """Update `title` value."""
        self._title_label.setText(value)

    @property
    def message(self) -> str:  # noqa: D401
        """Main message of the notification."""
        return self._title_label.text()

    @message.setter
    def message(self, value: str) -> None:
        """Update `message` value."""
        self._title_label.setText(value)

    @property
    def button_text(self) -> str:  # noqa: D401
        """Text of the confirmation button."""
        return self._accept_button.text()

    @button_text.setter
    def button_text(self, value: str) -> None:
        """Update `button_text` value."""
        self._accept_button.setText(value)


class EnterpriseRepoSetDomainPage(BaseSettingPage):  # pylint: disable=missing-class-docstring

    def __init__(self, parent: typing.Optional[QWidget] = None) -> None:
        """Initialize new :class:`~EnterpriseRepoSetDomainPage` instance."""
        super().__init__(AnacondaAPI(), ENTERPRISE_SET_DOMAIN_TEXT_CONTAINER, parent=parent)
        self.api_url_config_option = 'enterprise_4_repo_api_url'
        self.input_line.setPlaceholderText('http(s)://example.com')
        self.button_apply.clicked.connect(self.set_domain)


class EnterpriseRepoLoginPage(BaseLoginPage):  # pylint: disable=missing-class-docstring

    def __init__(self, parent: typing.Optional[QWidget] = None) -> None:
        """Initialize new :class:`~EnterpriseRepoLoginPage` instance."""
        super().__init__(AnacondaAPI(), ENTERPRISE_LOGIN_TEXT_CONTAINER, parent=parent)
        self.api_url_config_option = 'enterprise_4_repo_api_url'
        self.brand = AnacondaBrand.ENTERPRISE_EDITION
        self.text_login.setValidator(USER_RE_VALIDATOR)

    def create_login_data(self):
        """
        Backup actual .condarc and replace it with .condarc duplicate with empty channels list.
        """
        self.api.create_login_data()



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""
This folder contains image files bundled with Anaconda Navigator package.

This folder is defined as a python module so that some convenience global
variables can be defined.
"""

from __future__ import absolute_import, division, print_function

import os.path as osp


IMAGE_PATH = osp.dirname(osp.realpath(__file__))
LOGO_PATH = osp.join(IMAGE_PATH, 'logos')
ICONS_PATH = osp.join(IMAGE_PATH, 'icons')


# --- Anaconda Logo
# -----------------------------------------------------------------------------

ANACONDA_LOGO = osp.join(IMAGE_PATH, 'anaconda-logo.svg')
ANACONDA_LOGO_WHITE = osp.join(IMAGE_PATH, 'anaconda-logo-white.svg')
ANACONDA_NAVIGATOR_LOGO = osp.join(IMAGE_PATH, 'anaconda-navigator-logo.svg')
ANACONDA_ICON_256_PATH = osp.join(IMAGE_PATH, 'anaconda-icon-256x256.png')
ANACONDA_TEAM_EDITION_LOGIN_LOGO = osp.join(LOGO_PATH, 'Anaconda_Logo_RGB_Team_Edition.png')
ANACONDA_ORG_EDITION_LOGIN_LOGO = osp.join(LOGO_PATH, 'Anaconda_Logo_RGB_Org.png')
ANACONDA_COMMERCIAL_EDITION_LOGIN_LOGO = osp.join(LOGO_PATH, 'Anaconda_Logo_RGB_Commercial_Edition.png')
ANACONDA_ENTERPRISE_EDITION_LOGIN_LOGO = osp.join(LOGO_PATH, 'Anaconda_Logo_RGB_Enterprise_Edition.png')
ANACONDA_NUCLEUS_LOGIN_LOGO = osp.join(LOGO_PATH, 'Anaconda_Logo_RGB_Nucleus.png')

# NOTE: Check copyright on this image
VIDEO_ICON_PATH = osp.join(IMAGE_PATH, 'default-content.png')
CLOSE_DIALOG_ICON_PATH = osp.join(ICONS_PATH, 'close-icon.svg')
REFRESH_ICON_PATH = osp.join(ICONS_PATH, 'refresh.svg')
EXCLAMATION_CIRCLE_PATH = osp.join(ICONS_PATH, 'exclamation-circle.svg')
BLOCK_CLOSE_ICON_PATH = osp.join(ICONS_PATH, 'block-close.svg')

# --- Application icons
# -----------------------------------------------------------------------------
GLUEVIZ_ICON_1024_PATH = osp.join(IMAGE_PATH, 'glueviz-icon-1024x1024.png')
NOTEBOOK_ICON_1024_PATH = osp.join(IMAGE_PATH, 'jupyter-icon-1024x1024.png')
ORANGE_ICON_1024_PATH = osp.join(IMAGE_PATH, 'orange-icon-1024x1024.png')
QTCONSOLE_ICON_1024_PATH = osp.join(IMAGE_PATH, 'qtconsole-icon-1024x1024.png')
SPYDER_ICON_1024_PATH = osp.join(IMAGE_PATH, 'spyder-icon-1024x1024.png')
RODEO_ICON_1024_PATH = osp.join(IMAGE_PATH, 'rodeo-icon-1024x1024.png')
VEUSZ_ICON_1024_PATH = osp.join(IMAGE_PATH, 'veusz-icon-1024x1024.png')
RSTUDIO_ICON_1024_PATH = osp.join(IMAGE_PATH, 'rstudio-icon-1024x1024.png')
JUPYTERLAB_ICON_1024_PATH = osp.join(IMAGE_PATH, 'jupyterlab-icon-1024x1024.png')
VSCODE_ICON_1024_PATH = osp.join(IMAGE_PATH, 'vscode-icon-1024x1024.png')
PYCHARM_ICON_1024_PATH = osp.join(IMAGE_PATH, 'pycharm-icon.png')
PYCHARM_CE_ICON_1024_PATH = osp.join(IMAGE_PATH, 'pycharm-ce-icon.png')
DATALORE_ICON_1024_PATH = osp.join(IMAGE_PATH, 'datalore-icon.png')
QTCREATOR_ICON_1024_PATH = osp.join(IMAGE_PATH, 'qtcreator-icon-1024x1024.png')
IBM_WATSON_ICON_PATH = osp.join(IMAGE_PATH, 'ibm-watson-studio.png')

# --- Spinners
# -----------------------------------------------------------------------------
# http://preloaders.net/en/circular
SPINNER_16_PATH = osp.join(IMAGE_PATH, 'spinner-16x16.gif')
SPINNER_32_PATH = osp.join(IMAGE_PATH, 'spinner-32x32.gif')
SPINNER_GREEN_16_PATH = osp.join(IMAGE_PATH, 'spinner-green-16x16.gif')
SPINNER_WHITE_16_PATH = osp.join(IMAGE_PATH, 'spinner-white-16x16.gif')

# Conda Package Manager Table icons
# -----------------------------------------------------------------------------
MANAGER_INSTALLED = osp.join(IMAGE_PATH, 'icons', 'check-box-checked-active.svg')
MANAGER_NOT_INSTALLED = osp.join(IMAGE_PATH, 'icons', 'check-box-blank.svg')
MANAGER_ADD = osp.join(IMAGE_PATH, 'icons', 'mark-install.svg')
MANAGER_REMOVE = osp.join(IMAGE_PATH, 'icons', 'mark-remove.svg')
MANAGER_DOWNGRADE = osp.join(IMAGE_PATH, 'icons', 'mark-downgrade.svg')
MANAGER_UPGRADE = osp.join(IMAGE_PATH, 'icons', 'mark-upgrade.svg')
MANAGER_UPGRADE_ARROW = osp.join(IMAGE_PATH, 'icons', 'update-app-active.svg')
MANAGER_SPACER = osp.join(IMAGE_PATH, 'conda-manager-spacer.svg')
WARNING_ICON = osp.join(IMAGE_PATH, 'icons', 'warning-active.svg')
INFO_ICON = osp.join(IMAGE_PATH, 'icons', 'info-active.svg')
PYTHON_LOGO = osp.join(IMAGE_PATH, 'python-logo.svg')



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""
Workers and manager for running long processes in threads without blocking GUI.
"""

from collections import deque
import contextlib
import os
import sys
from qtpy.QtCore import QByteArray, QObject, QProcess, QThread, QTimer, Signal  # pylint: disable=no-name-in-module
from anaconda_navigator.utils.py3compat import PY2, to_text_string


WIN = os.name == 'nt'


def handle_qbytearray(obj, encoding):
    """Qt/Python2/3 compatibility helper."""
    if isinstance(obj, QByteArray):
        obj = obj.data()

    return to_text_string(obj, encoding=encoding)


class PythonWorker(QObject):
    """
    Generic python worker for running python code on threads.

    For running processes (via QProcess) use the ProcessWorker.
    """
    sig_started = Signal(object)
    sig_partial = Signal(object, object, object)  # worker, stdout, stderr
    sig_finished = Signal(object, object, object)  # worker, stdout, stderr

    def __init__(self, func, args, kwargs):
        """Generic python worker for running python code on threads."""
        super().__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._is_finished = False
        self._started = False

    def is_finished(self):
        """Return True if worker status is finished otherwise return False."""
        return self._is_finished

    def start(self):
        """Start the worker (emits sig_started signal with worker as arg)."""
        if not self._started:
            self.sig_started.emit(self)
            self._started = True

    def terminate(self):
        """Mark the worker as finished."""
        self._is_finished = True

    def _start(self):
        """Start process worker for given method args and kwargs."""
        error = None
        output = None

        try:
            output = self.func(*self.args, **self.kwargs)
        except Exception as err:  # pylint: disable=broad-except
            error = err

        if not self._is_finished:
            self.sig_finished.emit(self, output, error)
        self._is_finished = True


class DummyWorker(QObject):  # pylint: disable=too-few-public-methods
    """Process worker based on a QProcess for non blocking UI."""

    sig_started = Signal(object)
    sig_partial = Signal(object, object, object)
    sig_finished = Signal(object, object, object)


class ProcessWorker(QObject):  # pylint: disable=too-many-instance-attributes
    """Process worker based on a QProcess for non blocking UI."""

    sig_started = Signal(object)
    sig_partial = Signal(object, object, object)
    sig_finished = Signal(object, object, object)

    def __init__(self, cmd_list, environ=None):
        """
        Process worker based on a QProcess for non blocking UI.

        Parameters
        ----------
        cmd_list : list of str
            Command line arguments to execute.
        environ : dict
            Process environment,
        """
        super().__init__()
        self._result = None
        self._cmd_list = cmd_list
        self._fired = False
        self._communicate_first = False
        self._partial_stdout = None
        self._partial_stderr = None
        self._started = False

        self._timer = QTimer()
        self._process = QProcess()
        self._set_environment(environ)

        self._timer.setInterval(150)
        self._timer.timeout.connect(self._communicate)
        self._process.readyReadStandardOutput.connect(self._partial)

    @staticmethod
    def _get_encoding():
        """Return the encoding/codepage to use."""
        enco = 'utf-8'

        #  Currently only cp1252 is allowed?
        if WIN:
            import ctypes  # pylint: disable=import-outside-toplevel
            codepage = to_text_string(ctypes.cdll.kernel32.GetACP())
            enco = 'cp' + codepage

        return enco

    def _set_environment(self, environ):
        """Set the environment on the QProcess."""
        if environ:
            q_environ = self._process.processEnvironment()
            for k, v in environ.items():
                q_environ.insert(k, v)
            self._process.setProcessEnvironment(q_environ)

    def _partial(self):
        """Callback for partial output."""
        raw_stdout = self._process.readAllStandardOutput()
        stdout = handle_qbytearray(raw_stdout, self._get_encoding())

        if self._partial_stdout is None:
            self._partial_stdout = stdout
        else:
            self._partial_stdout += stdout

        # NOTE: use the piece or the cummulative?
        self.sig_partial.emit(self, stdout, None)

    def _communicate(self):
        """Callback for communicate."""
        if (not self._communicate_first and
                self._process.state() == QProcess.NotRunning):
            self.communicate()
        elif self._fired:
            self._timer.stop()

    def communicate(self):
        """Retrieve information."""
        self._communicate_first = True
        self._process.waitForFinished()

        enco = self._get_encoding()
        if self._partial_stdout is None:
            raw_stdout = self._process.readAllStandardOutput()
            stdout = handle_qbytearray(raw_stdout, enco)
        else:
            stdout = self._partial_stdout

        if self._partial_stderr is None:
            raw_stderr = self._process.readAllStandardError()
            stderr = handle_qbytearray(raw_stderr, enco)
        else:
            stderr = self._partial_stderr

        if PY2:
            stdout = stdout.decode()
            stderr = stderr.decode()

        result = [stdout, stderr]
        self._result = result

        if not self._fired:
            self.sig_finished.emit(self, result[0], result[-1])

        self._fired = True

        return result

    def close(self):
        """Close the running process."""
        self._process.close()

    def is_finished(self):
        """Return True if worker has finished processing."""
        return self._process.state() == QProcess.NotRunning and self._fired

    def _start(self):
        """Start process."""
        if not self._fired:
            # print(self._cmd_list)
            self._partial_ouput = None  # pylint: disable=attribute-defined-outside-init
            if self._cmd_list:
                self._process.start(self._cmd_list[0], self._cmd_list[1:])
            self._timer.start()

    def terminate(self):
        """Terminate running processes."""
        if self._process.state() == QProcess.Running:
            with contextlib.suppress(BaseException):
                self._process.terminate()
        self._fired = True

    def write(self, data):  # pylint: disable=missing-function-docstring
        if self._started:
            self._process.write(data)

    def start(self):
        """Start worker."""
        if not self._started:
            self.sig_started.emit(self)
            self._started = True


class WorkerManager(QObject):  # pylint: disable=too-many-instance-attributes
    """Spyder Worker Manager for Generic Workers."""

    def __init__(self, max_threads=10):
        """Spyder Worker Manager for Generic Workers."""
        super().__init__()
        self._queue = deque()
        self._queue_workers = deque()
        self._threads = []
        self._workers = []
        self._timer = QTimer()
        self._timer_worker_delete = QTimer()
        self._running_threads = 0
        self._max_threads = max_threads

        # Keeps references to old workers
        # Needed to avoid C++/python object errors
        self._bag_collector = deque()

        self._timer.setInterval(333)
        self._timer.timeout.connect(self._start)
        self._timer_worker_delete.setInterval(5000)
        self._timer_worker_delete.timeout.connect(self._clean_workers)

    def _clean_workers(self):
        """Dereference workers in workers bag periodically."""
        while self._bag_collector:
            self._bag_collector.popleft()
        self._timer_worker_delete.stop()

    def _start(self, worker=None):
        """Start threads and check for inactive workers."""
        if worker:
            self._queue_workers.append(worker)

        if self._queue_workers and self._running_threads < self._max_threads:
            self._running_threads += 1
            worker = self._queue_workers.popleft()
            thread = QThread()
            if isinstance(worker, PythonWorker):
                worker.moveToThread(thread)
                worker.sig_finished.connect(thread.quit)
                thread.started.connect(worker._start)  # pylint: disable=protected-access
                thread.start()
            elif isinstance(worker, ProcessWorker):
                thread.quit()
                worker._start()  # pylint: disable=protected-access
            self._threads.append(thread)
        else:
            self._timer.start()

        if self._workers:
            for w in self._workers:
                if w.is_finished():
                    self._bag_collector.append(w)
                    self._workers.remove(w)

        if self._threads:
            for t in self._threads:
                if t.isFinished():
                    self._threads.remove(t)
                    self._running_threads -= 1

        if len(self._threads) == 0 and len(self._workers) == 0:
            self._timer.stop()
            self._timer_worker_delete.start()

    def create_python_worker(self, func, *args, **kwargs):
        """Create a new python worker instance."""
        worker = PythonWorker(func, args, kwargs)
        self._create_worker(worker)
        return worker

    def create_process_worker(self, cmd_list, environ=None):
        """Create a new process worker instance."""
        worker = ProcessWorker(cmd_list, environ=environ)
        self._create_worker(worker)
        return worker

    def _create_worker(self, worker):
        """Common worker setup."""
        worker.sig_started.connect(self._start)
        self._workers.append(worker)


# --- Local testing
# -----------------------------------------------------------------------------
def ready_print(worker, output, error):  # pragma: no cover
    """Print worker output for tests."""
    print(worker, output, error)  # spyder: test-skip


def sleeping_func(arg, secs=10, result_queue=None):
    """This methods illustrates how the workers can be used."""
    import time  # pylint: disable=import-outside-toplevel
    time.sleep(secs)
    if result_queue is not None:
        result_queue.put(arg)
    else:
        return arg
    return None


def local_test():  # pragma: no cover
    """Main local test."""
    from spyder.utils.qthelpers import qapplication  # pylint: disable=import-error,import-outside-toplevel
    app = qapplication()
    wm = WorkerManager(max_threads=3)
    for i in range(7):
        worker = wm.create_python_worker(
            sleeping_func, f'BOOM! {i}', secs=5
        )
        worker.sig_finished.connect(ready_print)
        worker.start()
    worker = wm.create_python_worker(sleeping_func, 'BOOM!', secs=5)
    worker.sig_finished.connect(ready_print)
    worker.start()

    worker = wm.create_process_worker(['conda', 'info', '--json'])
    worker.sig_finished.connect(ready_print)
    worker.start()

    sys.exit(app.exec_())


if __name__ == '__main__':  # pragma: no cover
    local_test()



# -*- coding: utf-8 -*-

"""Type definitions for the :mod:`~anaconda_navigator.utils.conda`."""

from __future__ import annotations

__all__ = ()

import typing

if typing.TYPE_CHECKING:
    import typing_extensions

    CondaInfoOutput = typing_extensions.TypedDict('CondaInfoOutput', {
        'GID': int,
        'UID': int,
        'active_prefix': typing.Optional[str],
        'active_prefix_name': typing.Optional[str],
        'av_data_dir': str,
        'av_metadata_url_base': typing.Optional[str],
        'channels': typing.List[str],
        'conda_build_version': str,
        'conda_env_version': str,
        'conda_location': str,
        'conda_prefix': str,
        'conda_private': bool,
        'conda_shlvl': int,
        'conda_version': str,
        'config_files': typing.List[str],
        'default_prefix': str,
        'env_vars': dict,
        'envs': typing.List[str],
        'envs_dirs': typing.List[str],
        'netrc_file': typing.Optional[str],
        'offline': bool,
        'pkgs_dirs': typing.List[str],
        'platform': str,
        'python_version': str,
        'rc_path': str,
        'requests_version': str,
        'root_prefix': str,
        'root_writable': bool,
        'site_dirs': typing.List[str],
        'sys.executable': str,
        'sys.prefix': str,
        'sys.version': str,
        'sys_rc_path': str,
        'user_agent': str,
        'user_rc_path': str,
        'virtual_pkgs': typing.List[typing.List[str]],
    })

    class CondaErrorOutput(typing_extensions.TypedDict):
        """Common structure of the error reported by Conda."""

        caused_by: str
        error: str
        exception_name: str
        exception_type: str
        message: str

    class CondaValidationErrorOutput(CondaErrorOutput):
        """Error body of the :code:`CustomValidationError` reported by Conda."""

        parameter_name: str
        parameter_value: str
        source: str



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Base configuration management."""

import os
import os.path as osp
import sys
from anaconda_navigator.utils import encoding


# -----------------------------------------------------------------------------
# --- Configuration paths
# -----------------------------------------------------------------------------

SUBFOLDER = os.path.join('.anaconda', 'navigator')


def get_home_dir():
    """Return user home directory."""
    path = None
    if os.name == 'nt':
        path_env_vars = ('HOME', 'APPDATA', 'USERPROFILE', 'TMP')
        # prefer locations that already have .anaconda/navigator folders
        for env_var in path_env_vars:
            path = os.path.join(encoding.to_unicode_from_fs(os.environ.get(env_var, '')), SUBFOLDER)
            if osp.isdir(path):
                break
            path = None

        # did not find .anaconda/navigator folder; find first existing variable
        if not path:
            for env_var in path_env_vars:
                # os.environ.get() returns a raw byte string which needs to be
                # decoded with the codec that the OS is using
                # to represent environment variables.
                path = encoding.to_unicode_from_fs(os.environ.get(env_var, ''))
                if osp.isdir(path):
                    break
                path = None

    else:
        try:
            # expanduser() returns a raw byte string
            # which needs to be decoded with the codec
            # that the OS is using to represent file paths.
            path = encoding.to_unicode_from_fs(osp.expanduser('~'))
        except Exception:  # pylint: disable=broad-except
            path = ''
    if path:
        return path
    raise RuntimeError('Please define environment variable $HOME')


def get_conf_path(filename=None):
    """Return absolute path for configuration file with specified filename."""
    conf_dir = osp.join(get_home_dir(), SUBFOLDER)
    if not osp.isdir(conf_dir):
        os.makedirs(conf_dir)
    if filename is None:
        return conf_dir
    return osp.join(conf_dir, filename)


def is_ubuntu():
    """Detect if we are running in an Ubuntu-based distribution."""
    if sys.platform.startswith('linux') and os.path.isfile('/etc/lsb-release'):
        with open('/etc/lsb-release', 'r') as f:  # pylint: disable=invalid-name,unspecified-encoding
            release_info = f.read()
        return 'ubuntu' in release_info.lower()
    return False


def is_gtk_desktop():
    """Detect if we are running in a Gtk-based desktop."""
    if sys.platform.startswith('linux'):
        xdg_desktop = os.environ.get('XDG_CURRENT_DESKTOP', '')
        if xdg_desktop:
            gtk_desktops = ['Unity', 'GNOME', 'XFCE']
            return any(xdg_desktop.startswith(d) for d in gtk_desktops)
    return False
